from torchok.models.heads.base import AbstractHead
from torchok.models.heads import classification
from torchok.models.heads import detection
from torchok.models.heads import representation
from torchok.models.heads import segmentation


__all__ = [
    'AbstractHead',
]
