from torchok.tasks.base import BaseTask
from torchok.tasks.classification import ClassificationTask
# from src.tasks.moby import 
# from src.tasks.segmentation import 
# from src.tasks.representation import 
